@@ -268,6 +268,22 @@ pub struct GroupConfig {
     #[arg(long("max"), value_name("BYTES"))]
     pub max_size: Option<FileLen>,
 
+    /// Maximum prefix size to check in bytes
+    ///
+    /// Units like KB, KiB, MB, MiB, GB, GiB are supported.
+    ///
+    /// Default: 16KiB for hard drives, 4KiB for SSDs
+    #[arg(long("max-prefix-size"), value_name("BYTES"))]
+    pub max_prefix_size: Option<FileLen>,
+
+    /// Maximum suffix size to check in bytes
+    ///
+    /// Units like KB, KiB, MB, MiB, GB, GiB are supported.
+    ///
+    /// Default: 16KiB for hard drives, 4KiB for SSDs
+    #[arg(long("max-suffix-size"), value_name("BYTES"))]
+    pub max_suffix_size: Option<FileLen>,
+
     /// Include only file names matched fully by any of the given patterns.
     #[arg(long = "name", value_name("PATTERN"))]
     pub name_patterns: Vec<String>,
@@ -292,6 +308,19 @@ pub struct GroupConfig {
     #[arg(value_enum, long, value_name = "NAME", default_value = "metro")]
     pub hash_fn: HashFn,
 
+    /// Skip the full contents hash step entirely.
+    ///
+    /// This is a *dangerous* option - there is a significantly increased risk of false positives,
+    /// especially on smaller files or files with significant amounts of shared content that
+    /// wouldn't be caught by prefix/suffix checks.
+    ///
+    /// It's recommended to use this with increased `--max-prefix-size` and `--max-suffix-size`
+    /// options, to minimize false positives.
+    ///
+    /// Note: This option currently does nothing if you specify a `--transform` command
+    #[arg(long)]
+    pub skip_content_hash: bool,
+
     /// Enable caching of file hashes.
     ///
     /// Caching can significantly speed up subsequent runs of `fclones group` by avoiding
@@ -665,14 +694,13 @@ pub enum Command {
     /// The list of groups earlier produced by `fclones group` should be submitted
     /// on the standard input.
     ///
-    /// Unless `--soft` is specified, hard links are created for links within
-    /// the same file system. Soft links are always created to link files between
-    /// different file systems.
+    /// Unless `--soft` is specified, hard links are created.
+    /// Creating hard links to files on different mount points is expected to fail.
     Link {
         #[clap(flatten)]
         config: DedupeConfig,
 
-        /// Create soft links.
+        /// Create soft (symbolic) links.
         #[arg(short, long)]
         soft: bool,
     },
