@@ -6,6 +6,7 @@ use std::collections::BTreeMap;
 use std::collections::HashMap;
 use std::env::{args_os, current_dir};
 use std::ffi::{OsStr, OsString};
+use std::fmt::Debug;
 use std::fs::File;
 use std::hash::Hash;
 use std::io;
@@ -36,6 +37,7 @@ use crate::log::{Log, LogExt, ProgressBarLength};
 use crate::path::Path;
 use crate::phase::{Phase, Phases};
 use crate::report::{FileStats, ReportHeader, ReportWriter};
+use crate::rlimit::RLIMIT_OPEN_FILES;
 use crate::selector::PathSelector;
 use crate::semaphore::Semaphore;
 use crate::walk::Walk;
@@ -239,10 +241,72 @@ impl<F> FileGroup<F> {
     pub fn total_size(&self) -> FileLen {
         self.file_len * self.file_count() as u64
     }
+
+    /// Maps the list of files in the group.
+    /// Preserves the group file len and hash.
+    pub fn map<R>(self, f: impl Fn(F) -> R) -> FileGroup<R> {
+        FileGroup {
+            file_len: self.file_len,
+            file_hash: self.file_hash,
+            files: self.files.into_iter().map(f).collect(),
+        }
+    }
+
+    /// Tries to map each file by given fallible function.
+    /// Does not stop processing on the first failure.
+    /// If mapping any of the files fails, then returns a vector of errors.
+    pub fn try_map_all<R: Debug, E: Debug>(
+        self,
+        f: impl Fn(F) -> Result<R, E>,
+    ) -> Result<FileGroup<R>, Vec<E>> {
+        let (ok, err): (Vec<_>, Vec<_>) = self.files.into_iter().map(f).partition(Result::is_ok);
+        if err.is_empty() {
+            Ok(FileGroup {
+                file_len: self.file_len,
+                file_hash: self.file_hash,
+                files: ok.into_iter().map(Result::unwrap).collect(),
+            })
+        } else {
+            Err(err.into_iter().map(Result::unwrap_err).collect())
+        }
+    }
+
+    /// Flat maps the list of files in the group.
+    /// Preserves the group file len and hash.
+    pub fn flat_map<R, I>(self, f: impl Fn(F) -> I) -> FileGroup<R>
+    where
+        I: IntoIterator<Item = R>,
+    {
+        FileGroup {
+            file_len: self.file_len,
+            file_hash: self.file_hash,
+            files: self.files.into_iter().flat_map(f).collect(),
+        }
+    }
+
+    /// Splits the group into one or more groups based on the key function applied to each file.
+    /// Files with the same key are placed in the same group.
+    /// The key is computed only once per item.
+    /// File len and file hash are preserved.
+    pub fn partition_by_key<K: Eq + Hash>(self, key_fn: impl Fn(&F) -> K) -> Vec<FileGroup<F>> {
+        let mut groups = HashMap::new();
+        for f in self.files {
+            let key = key_fn(&f);
+            groups.entry(key).or_insert_with(Vec::new).push(f);
+        }
+        groups
+            .into_values()
+            .map(|files| FileGroup {
+                file_len: self.file_len,
+                file_hash: self.file_hash.clone(),
+                files,
+            })
+            .collect()
+    }
 }
 
-#[cfg(test)]
-impl<F: AsRef<Path>> FileGroup<F> {
+impl<F: AsRef<Path> + core::fmt::Debug> FileGroup<F> {
+    #[cfg(test)]
     fn paths(&self) -> Vec<Path> {
         self.files.iter().map(|f| f.as_ref().clone()).collect_vec()
     }
@@ -572,6 +636,7 @@ where
                     // when the pool has only one thread.
                     let hash_fn: &HashFn<'static> = unsafe { std::mem::transmute(hash_fn) };
                     thread_pool.spawn_fifo(move || {
+                        let _open_files_guard = RLIMIT_OPEN_FILES.clone().access_owned();
                         let old_hash = fg[0].file_hash.clone();
                         if let Some(hash) = hash_fn((&mut fg[0].file_info, old_hash)) {
                             for mut f in fg {
@@ -809,9 +874,16 @@ fn update_file_locations(ctx: &GroupCtx<'_>, groups: &mut (impl FileCollection +
     let err_counters = atomic_counter_vec(ctx.devices.len());
     groups.for_each_mut(|fi| {
         let device: &DiskDevice = &ctx.devices[fi.get_device_index()];
-        if device.disk_type != sysinfo::DiskType::SSD {
+        if device.disk_kind != sysinfo::DiskKind::SSD {
             if let Err(e) = fi.fetch_physical_location() {
-                handle_fetch_physical_location_err(ctx, &err_counters, fi, e)
+                // Do not print a notice about slower access when fetching file extents has
+                // failed because a file vanished -- now it will never be accessed anyhow.
+                const ENOENT_NO_SUCH_FILE: i32 = 2;
+                if e.raw_os_error()
+                    .map_or(true, |err| err != ENOENT_NO_SUCH_FILE)
+                {
+                    handle_fetch_physical_location_err(ctx, &err_counters, fi, e)
+                }
             }
         }
         progress.inc(1)
@@ -999,7 +1071,10 @@ fn group_by_suffix(
     let mut groups = groups;
     sort_files_by_id(&mut groups);
 
-    let suffix_len = suffix_len(&ctx.devices, flat_iter(&groups));
+    let suffix_len = ctx
+        .config
+        .max_suffix_size
+        .unwrap_or_else(|| suffix_len(&ctx.devices, flat_iter(&groups)));
     let suffix_threshold = suffix_threshold(&ctx.devices, flat_iter(&groups));
     let pre_filter =
         |g: &FileGroup<FileInfo>| g.file_len >= suffix_threshold && g.unique_count() > 1;
@@ -1153,10 +1228,17 @@ pub fn group_files(config: &GroupConfig, log: &dyn Log) -> Result<Vec<FileGroup<
             let size_groups = group_by_size(&ctx, matching_files);
             let mut size_groups_pruned = remove_same_files(&ctx, size_groups);
             update_file_locations(&ctx, &mut size_groups_pruned);
-            let prefix_len = prefix_len(&ctx.devices, flat_iter(&size_groups_pruned));
+            let prefix_len = ctx
+                .config
+                .max_prefix_size
+                .unwrap_or_else(|| prefix_len(&ctx.devices, flat_iter(&size_groups_pruned)));
             let prefix_groups = group_by_prefix(&ctx, prefix_len, size_groups_pruned);
             let suffix_groups = group_by_suffix(&ctx, prefix_groups);
-            group_by_contents(&ctx, prefix_len, suffix_groups)
+            if !ctx.config.skip_content_hash {
+                group_by_contents(&ctx, prefix_len, suffix_groups)
+            } else {
+                suffix_groups
+            }
         }
     };
     groups.par_sort_by_key(|g| Reverse((g.file_len, g.file_hash.u128_prefix())));
@@ -1242,7 +1324,7 @@ mod test {
 
     use crate::log::StdLog;
     use rand::seq::SliceRandom;
-    use sysinfo::DiskType;
+    use sysinfo::DiskKind;
 
     use crate::path::Path;
     use crate::util::test::*;
@@ -1484,7 +1566,7 @@ mod test {
     #[test]
     fn test_rehash_processes_files_in_location_order_on_hdd() {
         let thread_count = 2;
-        let devices = DiskDevices::single(DiskType::HDD, thread_count);
+        let devices = DiskDevices::single(DiskKind::HDD, thread_count);
         let count = 1000;
         let mut input = Vec::with_capacity(count);
         for i in 0..count {
@@ -1922,6 +2004,75 @@ mod test {
         )
     }
 
+    #[test]
+    fn partition() {
+        let fg = FileGroup {
+            file_len: FileLen::from(1u64),
+            file_hash: FileHash::from(1u128),
+            files: vec!["a1", "b1", "a2", "b2", "b3"],
+        };
+        let mut partitions = fg.partition_by_key(|f| f.chars().next().unwrap());
+        assert_eq!(partitions.len(), 2);
+        partitions.sort_by_key(|p| p.files.len());
+        assert_eq!(partitions[0].files, vec!["a1", "a2"]);
+        assert_eq!(partitions[1].files, vec!["b1", "b2", "b3"]);
+    }
+
+    #[test]
+    fn map() {
+        let fg = FileGroup {
+            file_len: FileLen::from(1u64),
+            file_hash: FileHash::from(1u128),
+            files: vec!["a", "b"],
+        };
+        let fg = fg.map(|f| format!("{f}.txt"));
+        assert_eq!(fg.files, vec![String::from("a.txt"), String::from("b.txt")]);
+    }
+
+    #[test]
+    fn try_map_all_happy_path() {
+        let fg = FileGroup {
+            file_len: FileLen::from(1u64),
+            file_hash: FileHash::from(1u128),
+            files: vec!["a", "b"],
+        };
+        let fg = fg.try_map_all(|f| Result::<_, ()>::Ok(format!("{f}.txt")));
+        assert!(fg.is_ok());
+        assert_eq!(
+            fg.unwrap().files,
+            vec![String::from("a.txt"), String::from("b.txt")]
+        );
+    }
+
+    #[test]
+    fn try_map_all_errors() {
+        let fg = FileGroup {
+            file_len: FileLen::from(1u64),
+            file_hash: FileHash::from(1u128),
+            files: vec!["a", "b"],
+        };
+        let fg = fg.try_map_all(|f| Result::<(), _>::Err(format!("error {f}")));
+        assert!(fg.is_err());
+        assert_eq!(
+            fg.unwrap_err(),
+            vec![String::from("error a"), String::from("error b")]
+        );
+    }
+
+    #[test]
+    fn flat_map() {
+        let fg = FileGroup {
+            file_len: FileLen::from(1u64),
+            file_hash: FileHash::from(1u128),
+            files: vec!["a1", "b1", "a2", "b2", "b3"],
+        };
+        let fg = fg.flat_map(|f| if f.starts_with('a') { Some(f) } else { None });
+        assert_eq!(fg.files, vec!["a1", "a2"]);
+
+        let fg = fg.flat_map(|f| vec![f, f]);
+        assert_eq!(fg.files, vec!["a1", "a1", "a2", "a2"]);
+    }
+
     fn write_test_file(path: &PathBuf, prefix: &[u8], mid: &[u8], suffix: &[u8]) {
         let mut file = OpenOptions::new()
             .write(true)
